package com.chelvc.framework.base.apidoc;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.time.Duration;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.arsframework.apidoc.core.Parameter;
import com.arsframework.apidoc.core.ParameterAnalyser;
import com.chelvc.framework.base.jackson.EnumerationFormatSerializer;
import com.chelvc.framework.common.model.Enumeration;
import com.chelvc.framework.common.model.Paging;
import com.chelvc.framework.common.model.Period;
import com.chelvc.framework.common.model.Region;
import com.chelvc.framework.common.util.ObjectUtils;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import org.springframework.util.CollectionUtils;

/**
 * 自定义接口文档参数解析实现
 *
 * @author Woody
 * @date 2023/1/1
 */
public class CustomizeParameterAnalyser extends ParameterAnalyser {
    public CustomizeParameterAnalyser(Method method) {
        super(method);
    }

    /**
     * 判断参数是否是枚举响应参数
     *
     * @param parameter 参数对象实例
     * @return true/false
     */
    private boolean isResponseEnumerationParameter(Parameter parameter) {
        return !parameter.isInput() && Enumeration.class.isAssignableFrom(parameter.getOriginal());
    }

    /**
     * 判断参数是否需要枚举格式化
     *
     * @param parameter 参数对象实例
     * @param serialize Jackson序列化注解实例
     * @return true/false
     */
    private boolean isEnumerationFormatted(Parameter parameter, JsonSerialize serialize) {
        if (serialize == null) {
            return false;
        } else if (parameter.isMultiple()) {
            Class<?> serializer = ObjectUtils.ifNull(serialize, JsonSerialize::contentUsing);
            return serializer != null && EnumerationFormatSerializer.class.isAssignableFrom(serializer);
        }
        Class<?> serializer = ObjectUtils.ifNull(serialize, JsonSerialize::using);
        return serializer != null && EnumerationFormatSerializer.class.isAssignableFrom(serializer);
    }

    /**
     * 判断参数是否需要枚举格式化
     *
     * @param parameter 参数对象实例
     * @return true/false
     */
    private boolean isEnumerationFormatted(Parameter parameter) {
        if (!this.isResponseEnumerationParameter(parameter)) {
            return false;
        }
        JsonSerialize serialize = parameter.getOriginal().getAnnotation(JsonSerialize.class);
        return this.isEnumerationFormatted(parameter, serialize);
    }

    /**
     * 判断参数是否需要枚举格式化
     *
     * @param parameter 参数对象实例
     * @param field     参数字段实例
     * @return true/false
     */
    private boolean isEnumerationFormatted(Parameter parameter, Field field) {
        // 排除已经是枚举格式化参数场景，避免重复设置
        if (!this.isResponseEnumerationParameter(parameter) || this.isEnumerationFormatted(parameter)) {
            return false;
        }
        JsonSerialize serialize = field.getAnnotation(JsonSerialize.class);
        return this.isEnumerationFormatted(parameter, serialize);
    }

    /**
     * 初始化枚举参数格式化信息
     *
     * @param parameter 参数对象实例
     */
    private void initializeEnumerationFormation(Parameter parameter) {
        parameter.setType(Object.class);
        List<Parameter.Option> options = parameter.getOptions();
        Parameter.Option example = CollectionUtils.isEmpty(options) ? null : options.get(0);
        List<Parameter> fields = Arrays.asList(
                Parameter.builder().input(false).type(String.class).original(String.class)
                        .name("code").description("编码").options(options)
                        .example(ObjectUtils.ifNull(example, Parameter.Option::getKey)).build(),
                Parameter.builder().input(false).type(String.class).original(String.class)
                        .name("description").description("描述").options(options)
                        .example(ObjectUtils.ifNull(example, Parameter.Option::getValue)).build()
        );
        parameter.setFields(fields);
    }

    @Override
    protected Parameter field2parameter(boolean input, Object instance, Field field,
                                        Map<TypeVariable<?>, Type> variables, LinkedList<Class<?>> stack) {
        Parameter parameter = super.field2parameter(input, instance, field, variables, stack);
        if (this.isEnumerationFormatted(parameter, field)) {
            this.initializeEnumerationFormation(parameter);
        }
        return parameter;
    }

    @Override
    protected void afterInitializeParameter(Parameter parameter) {
        if (parameter.getOriginal() == Paging.class) {
            parameter.setType(Paging.class);
            parameter.setFields(null);
            parameter.setFormat("页码,分页大小");
            parameter.setExample("1,10");
        } else if (parameter.getOriginal() == Period.class) {
            parameter.setType(Period.class);
            parameter.setFields(null);
            parameter.setFormat("开始时间,结束时间");
            parameter.setExample("2023-01-01,2023-01-01 23:59:59");
        } else if (parameter.getOriginal() == Region.class) {
            parameter.setType(Integer.class);
            parameter.setFields(null);
            parameter.setExample("110000");
        } else if (parameter.getOriginal() == Duration.class) {
            parameter.setType(String.class);
            parameter.setFields(null);
            parameter.setExample("PT1S");
        } else if (this.isEnumerationFormatted(parameter)) {
            this.initializeEnumerationFormation(parameter);
        }
    }
}
