package com.chelvc.framework.base.interceptor;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.CompletionException;
import java.util.stream.Collectors;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.validation.ConstraintViolation;
import javax.validation.ConstraintViolationException;
import javax.validation.ValidationException;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.base.exception.FrameworkException;
import com.chelvc.framework.base.model.Result;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.exc.InvalidFormatException;
import com.google.common.collect.ImmutableMap;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.TypeMismatchException;
import org.springframework.boot.autoconfigure.web.servlet.error.AbstractErrorController;
import org.springframework.boot.web.error.ErrorAttributeOptions;
import org.springframework.boot.web.servlet.error.DefaultErrorAttributes;
import org.springframework.http.HttpStatus;
import org.springframework.http.converter.HttpMessageConversionException;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.stereotype.Controller;
import org.springframework.util.CollectionUtils;
import org.springframework.validation.BindException;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.web.HttpMediaTypeNotAcceptableException;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.MissingMatrixVariableException;
import org.springframework.web.bind.MissingPathVariableException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.bind.ServletRequestBindingException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestControllerAdvice;
import org.springframework.web.method.annotation.MethodArgumentTypeMismatchException;
import org.springframework.web.multipart.MultipartException;
import org.springframework.web.multipart.support.MissingServletRequestPartException;

/**
 * 异常处理拦截器
 *
 * @author Woody
 * @date 2023/1/1
 */
@Slf4j
@Controller
@RestControllerAdvice
public class GlobalExceptionInterceptor extends AbstractErrorController {
    private final ErrorAttributeOptions options = ErrorAttributeOptions.of(ErrorAttributeOptions.Include.values());

    public GlobalExceptionInterceptor() {
        super(new DefaultErrorAttributes(), Collections.emptyList());
    }

    @Override
    public String getErrorPath() {
        return null;
    }

    /**
     * 参数绑定结果格式化
     *
     * @param result 参数绑定结果
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(BindingResult result) {
        List<FieldError> errors = ObjectUtils.ifNull(result, BindingResult::getFieldErrors);
        if (CollectionUtils.isEmpty(errors)) {
            return Collections.emptyMap();
        }
        return errors.stream().collect(Collectors.toMap(FieldError::getField,
                e -> ObjectUtils.ifNull(e.getDefaultMessage(), StringUtils.EMPTY), (o, n) -> n));
    }

    /**
     * 参数校验异常格式化
     *
     * @param e 参数校验异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(ConstraintViolationException e) {
        Set<ConstraintViolation<?>> errors =
                ObjectUtils.ifNull(e, ConstraintViolationException::getConstraintViolations);
        if (CollectionUtils.isEmpty(errors)) {
            return Collections.emptyMap();
        }
        return errors.stream().filter(Objects::nonNull).collect(Collectors.toMap(
                error -> {
                    String path = String.valueOf(error.getPropertyPath());
                    int index = path.lastIndexOf('.');
                    return index > -1 ? path.substring(index + 1) : path;
                },
                error -> ObjectUtils.ifNull(error.getMessage(), StringUtils.EMPTY)
        ));
    }

    /**
     * 参数类型异常格式化
     *
     * @param e 参数类型异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(MethodArgumentTypeMismatchException e) {
        return ImmutableMap.of(e.getName(),
                ObjectUtils.ifNull(e.getRequiredType(), Class::getSimpleName, () -> StringUtils.EMPTY));
    }

    /**
     * 参数类型异常格式化
     *
     * @param e 参数校验异常
     * @return 参数名称/异常消息映射表
     */
    private Map<String, String> format(HttpMessageNotReadableException e) {
        Throwable cause = e.getCause();
        if (cause instanceof InvalidFormatException) {
            List<JsonMappingException.Reference> references = ((InvalidFormatException) cause).getPath();
            if (!CollectionUtils.isEmpty(references)) {
                return references.stream().collect(Collectors.toMap(
                        JsonMappingException.Reference::getFieldName, r -> "Invalid format"
                ));
            }
        }
        return Collections.emptyMap();
    }

    /**
     * 将异常转换成结果对象
     *
     * @param request Http请求对象
     * @param e       异常对象
     * @return 结果对象
     */
    private Result<?> error2result(HttpServletRequest request, Throwable e) {
        if (e instanceof CompletionException) {
            return this.error2result(request, e.getCause());
        } else if (e instanceof HttpRequestMethodNotSupportedException) {
            return Result.build(HttpStatus.METHOD_NOT_ALLOWED, null, e.getMessage());
        } else if (e instanceof HttpMediaTypeNotAcceptableException) {
            return Result.build(HttpStatus.NOT_ACCEPTABLE, null, e.getMessage());
        } else if (e instanceof HttpMediaTypeNotSupportedException) {
            return Result.build(HttpStatus.UNSUPPORTED_MEDIA_TYPE, null, e.getMessage());
        } else if (e instanceof HttpMessageNotReadableException) {
            Map<String, String> errors = this.format((HttpMessageNotReadableException) e);
            return CollectionUtils.isEmpty(errors) ? Result.build(HttpStatus.BAD_REQUEST) :
                    Result.build(HttpStatus.PRECONDITION_FAILED, errors);
        } else if (e instanceof MissingPathVariableException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ImmutableMap.of(
                    ((MissingPathVariableException) e).getVariableName(), e.getMessage()
            ));
        } else if (e instanceof MissingMatrixVariableException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ImmutableMap.of(
                    ((MissingMatrixVariableException) e).getVariableName(), e.getMessage()
            ));
        } else if (e instanceof MissingServletRequestParameterException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ImmutableMap.of(
                    ((MissingServletRequestParameterException) e).getParameterName(), e.getMessage()
            ));
        } else if (e instanceof BindException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format(((BindException) e).getBindingResult()));
        } else if (e instanceof ConstraintViolationException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format((ConstraintViolationException) e));
        } else if (e instanceof MethodArgumentNotValidException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED,
                    this.format(((MethodArgumentNotValidException) e).getBindingResult()));
        } else if (e instanceof MethodArgumentTypeMismatchException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, this.format((MethodArgumentTypeMismatchException) e));
        } else if (e instanceof MissingServletRequestPartException) {
            return Result.build(HttpStatus.PRECONDITION_FAILED, ImmutableMap.of(
                    ((MissingServletRequestPartException) e).getRequestPartName(), e.getMessage()
            ));
        } else if (e instanceof ValidationException || e instanceof TypeMismatchException
                || e instanceof MultipartException || e instanceof HttpMessageConversionException
                || e instanceof ServletRequestBindingException) {
            return Result.build(HttpStatus.BAD_REQUEST, null, e.getMessage());
        } else if (e instanceof FrameworkException) {
            FrameworkException exception = (FrameworkException) e;
            return Result.build(exception.getStatus(), exception.getData(), exception.getMessage());
        }
        boolean wrapping = SessionContextHolder.isResponseWrappingHeader(request);
        return Result.build(HttpStatus.INTERNAL_SERVER_ERROR, null, wrapping ? null : e.getMessage());
    }

    /**
     * 打印异常日志
     *
     * @param request   Http请求对象
     * @param status    Http状态枚举
     * @param exception 异常对象
     */
    private void log(@NonNull HttpServletRequest request, @NonNull HttpStatus status, @NonNull Exception exception) {
        String message = exception.getMessage();
        if (StringUtils.isEmpty(message)) {
            message = status.getReasonPhrase();
        }
        message = SessionContextHolder.getLoggingMessage(request, status, message);
        if (status.is4xxClientError()) {
            if (log.isDebugEnabled()) {
                log.warn(message, exception);
            } else {
                log.warn(message);
            }
        } else {
            log.error(message, exception);
        }
    }

    /**
     * 异常方法
     *
     * @param request  Http请求对象
     * @param response Http响应对象
     * @return 异常结果
     */
    @RequestMapping("${server.error.path:${error.path:/error}}")
    public Result<?> error(HttpServletRequest request, HttpServletResponse response) {
        HttpStatus status = this.getStatus(request);
        Map<String, Object> attributes = this.getErrorAttributes(request, this.options);
        String uri = (String) attributes.get("path");
        String message = (String) attributes.get("message");
        log.warn(SessionContextHolder.getLoggingMessage(uri, request.getMethod(), status.value(), message));
        response.setStatus(status.value());
        SessionContextHolder.initializeResponseWrappingHeader(response);
        return Result.build(status, null, message);
    }

    /**
     * 异常转换
     *
     * @param request   Http请求对象
     * @param response  Http响应对象
     * @param exception 异常对象
     * @return 异常结果
     */
    @ExceptionHandler(Exception.class)
    public Result<?> exception(HttpServletRequest request, HttpServletResponse response, Exception exception) {
        Result<?> result = this.error2result(request, exception);
        HttpStatus status = HttpUtils.getStatus(result.getCode());
        this.log(request, status, exception);
        response.setStatus(status.value());
        SessionContextHolder.initializeResponseWrappingHeader(response);
        return result;
    }
}