package com.chelvc.framework.base.jackson;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;

import com.chelvc.framework.base.annotation.Decimal;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.util.ObjectUtils;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.BeanProperty;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ser.ContextualSerializer;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;

/**
 * Jackson小数序格式化处理器
 *
 * @author Woody
 * @date 2023/1/1
 */
public class DecimalFormatSerializer extends StdSerializer<Number> implements ContextualSerializer {
    private Decimal annotation;
    private boolean enabled = true;

    public DecimalFormatSerializer() {
        super(Number.class);
    }

    /**
     * 获取小数保留位数
     *
     * @return 小数保留位数
     */
    private int getScale() {
        Integer scale = ObjectUtils.ifNull(this.annotation, Decimal::scale);
        return scale == null || scale < 1 ? Decimal.DEFAULT_SCALE : scale;
    }

    /**
     * 获取小数舍弃模式
     *
     * @return 小数舍弃模式
     */
    private RoundingMode getMode() {
        return ObjectUtils.ifNull(ObjectUtils.ifNull(this.annotation, Decimal::mode), RoundingMode.HALF_UP);
    }

    /**
     * 是否去掉多余的0
     *
     * @return true/false
     */
    private boolean isTrimZeros() {
        return ObjectUtils.ifNull(ObjectUtils.ifNull(this.annotation, Decimal::trimZeros), false);
    }

    /**
     * 判断是否受版本控制
     *
     * @param annotation 小数格式化注解实例
     * @return true/false
     */
    private boolean isVersioning(Decimal annotation) {
        return ObjectUtils.isEmpty(annotation.versioning()) || SessionContextHolder.isVersion(annotation.versioning());
    }

    /**
     * 判断是否是小数属性
     *
     * @param property 属性对象
     * @return true/false
     */
    private boolean isDecimalProperty(BeanProperty property) {
        return Float.class.isAssignableFrom(property.getType().getRawClass())
                || Double.class.isAssignableFrom(property.getType().getRawClass())
                || BigDecimal.class.isAssignableFrom(property.getType().getRawClass());
    }

    /**
     * 获取属性小数注解实例
     *
     * @param property 属性对象
     * @return 小数注解实例
     */
    private Decimal getDecimalAnnotation(BeanProperty property) {
        return property.getAnnotation(Decimal.class);
    }

    @Override
    public void serialize(Number number, JsonGenerator generator, SerializerProvider provider)
            throws IOException {
        if (!this.enabled || number == null) {
            generator.writeObject(number);
        } else {
            BigDecimal decimal = number instanceof BigDecimal ? (BigDecimal) number : new BigDecimal(number.toString());
            decimal = decimal.setScale(this.getScale(), this.getMode());
            if (this.isTrimZeros()) {
                decimal = decimal.stripTrailingZeros();
            }
            generator.writeNumber(decimal.toPlainString());
        }
    }

    @Override
    public JsonSerializer<?> createContextual(SerializerProvider provider, BeanProperty property) {
        if (property == null) {
            return provider.getDefaultNullValueSerializer();
        }
        if (!this.isDecimalProperty(property)) {
            this.enabled = false;
        } else {
            this.annotation = this.getDecimalAnnotation(property);
            this.enabled = this.annotation == null || this.isVersioning(this.annotation);
        }
        return this;
    }
}
