package com.chelvc.framework.common.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.Objects;

import lombok.NonNull;

/**
 * 异常工具类
 *
 * @author Woody
 * @date 2023/1/1
 */
public final class ErrorUtils {
    private ErrorUtils() {
    }

    /**
     * 查找异常栈中根异常
     *
     * @param throwable 异常栈
     * @return 根异常
     */
    public static Throwable root(Throwable throwable) {
        while (throwable != null && throwable.getCause() != null) {
            throwable = throwable.getCause();
        }
        return throwable;
    }

    /**
     * 判断异常是否包含任意指定异常
     *
     * @param throwable 异常栈
     * @param types     异常类型数组
     * @return true/false
     */
    public static boolean isError(Throwable throwable, Class<? extends Throwable>[] types) {
        return types != null && types.length > 0 && isError(throwable, Arrays.asList(types));
    }

    /**
     * 判断异常是否包含任意指定异常
     *
     * @param throwable 异常栈
     * @param types     异常类型数组
     * @return true/false
     */
    public static boolean isError(Throwable throwable, Collection<Class<? extends Throwable>> types) {
        return ObjectUtils.nonEmpty(types)
                && types.stream().filter(Objects::nonNull).anyMatch(type -> contains(throwable, type));
    }

    /**
     * 判断异常栈是否包含制定类型的异常
     *
     * @param throwable 异常栈
     * @param type      异常类型
     * @return true/false
     */
    public static boolean contains(Throwable throwable, Class<? extends Throwable> type) {
        return lookup(throwable, type) != null;
    }

    /**
     * 从异常栈中查找制定类型的异常
     *
     * @param throwable 异常栈
     * @param type      异常类型
     * @param <T>       异常类型泛型
     * @return 异常对象
     */
    @SuppressWarnings("unchecked")
    public static <T extends Throwable> T lookup(Throwable throwable, @NonNull Class<T> type) {
        for (; throwable != null && !type.isAssignableFrom(throwable.getClass()); throwable = throwable.getCause()) ;
        return (T) throwable;
    }
}
