package com.chelvc.framework.rocketmq.consumer;

import java.lang.reflect.Type;
import java.time.Duration;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.rocketmq.annotation.RocketMQConsumer;
import com.chelvc.framework.rocketmq.config.RocketMQProperties;
import com.chelvc.framework.rocketmq.context.RocketMQContextHolder;
import com.chelvc.framework.rocketmq.interceptor.RocketMQListener;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.apis.ClientConfiguration;
import org.apache.rocketmq.client.apis.ClientException;
import org.apache.rocketmq.client.apis.ClientServiceProvider;
import org.apache.rocketmq.client.apis.consumer.FilterExpression;
import org.apache.rocketmq.client.apis.consumer.FilterExpressionType;
import org.apache.rocketmq.client.apis.consumer.SimpleConsumer;
import org.apache.rocketmq.client.apis.consumer.SimpleConsumerBuilder;
import org.apache.rocketmq.client.apis.message.MessageId;
import org.apache.rocketmq.client.apis.message.MessageView;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.context.ApplicationContext;
import org.springframework.core.env.Environment;

/**
 * 消息批量消费监听器容器
 *
 * @param <T> 监听消息对象类型
 * @author Woody
 * @date 2023/1/1
 */
@Slf4j
public class MultipleRocketMQListenerContainer<T> implements RocketMQListenerContainer<T> {
    private Type type;
    private String topic;
    private String group;
    private int quantity;
    private Duration duration;
    private Thread thread;
    private SimpleConsumer consumer;
    private RocketMQListener<T> listener;
    private volatile boolean running = true;

    /**
     * 接收消息
     *
     * @return 消息视图列表
     */
    private List<MessageView> receive() {
        try {
            return this.consumer.receive(this.quantity, this.duration);
        } catch (Throwable t) {
            log.error("Message receive failed[{}][{}]: {}", this.group, this.topic, t.getMessage(), t);
        }
        return Collections.emptyList();
    }

    /**
     * 批量消费消息
     *
     * @param messages 消息视图列表
     * @return true/false
     */
    @SuppressWarnings("unchecked")
    private boolean consume(List<MessageView> messages) {
        try {
            List<T> payloads = messages.stream()
                    .map(message -> (T) RocketMQContextHolder.deserialize(message, this.type))
                    .collect(Collectors.toList());
            this.listener.consume(payloads);
            return true;
        } catch (Throwable t) {
            String ids = messages.stream().map(MessageView::getMessageId).map(MessageId::toString)
                    .collect(Collectors.joining(","));
            log.error("Message consume failed[{}][{}]: {}, {}", this.group, this.topic, ids, t.getMessage(), t);
        }
        return false;
    }

    /**
     * 确认消息
     *
     * @param messages 消息视图列表
     */
    private void acknowledge(List<MessageView> messages) {
        for (MessageView message : messages) {
            try {
                this.consumer.ack(message);
            } catch (Throwable t) {
                String id = message.getMessageId().toString();
                log.error("Message acknowledge failed[{}][{}]: {}, {}", this.group, this.topic, id, t.getMessage(), t);
            }
        }
    }

    /**
     * 启动消费线程
     *
     * @return 消费线程实例
     */
    private Thread startup() {
        return ThreadUtils.run(() -> {
            while (this.running && !Thread.interrupted()) {
                // 批量接收消息
                List<MessageView> messages = this.receive();
                if (ObjectUtils.isEmpty(messages)) {
                    continue;
                }

                // 消费消息
                boolean success = this.consume(messages);

                // 回执消息
                if (success) {
                    this.acknowledge(messages);
                }
            }
        });
    }

    /**
     * 关闭消息消费者线程
     *
     * @param thread 消息消费者线程
     */
    private void shutdown(Thread thread) {
        if (thread != null) {
            thread.interrupt();
            ThreadUtils.join(thread);
        }
    }

    /**
     * 关闭消息消费者
     *
     * @param consumer 消息消费者实例
     */
    private void shutdown(SimpleConsumer consumer) {
        if (consumer != null) {
            try {
                consumer.close();
            } catch (Throwable t) {
                log.warn("Message consumer shutdown failed[{}][{}]: {}", this.group, this.topic, t.getMessage(), t);
            }
        }
    }

    @Override
    public void initialize(@NonNull ClientServiceProvider provider, @NonNull ClientConfiguration configuration,
                           @NonNull ApplicationContext applicationContext, @NonNull RocketMQListener<T> listener,
                           @NonNull RocketMQConsumer annotation, @NonNull RocketMQProperties properties) {
        // 参数配置有效性校验
        Class<?> clazz = AopProxyUtils.ultimateTargetClass(listener);
        AssertUtils.check(StringUtils.nonEmpty(annotation.topic()),
                "Message consumer topic must not be empty: " + clazz.getName());
        AssertUtils.check(StringUtils.nonEmpty(annotation.group()),
                "Message consumer group must not be empty: " + clazz.getName());
        AssertUtils.check(annotation.quantity() > 0, "Message consumer quantity must be greater than 0");
        AssertUtils.check(annotation.duration() > 0, "Message consumer duration must be greater than 0");
        AssertUtils.check(annotation.interval() > 0, "Message consumer interval must be greater than 0");
        AssertUtils.check(StringUtils.nonEmpty(annotation.tag()), "Message consumer tag must not be empty");

        // 初始化消费者实例
        Environment environment = applicationContext.getEnvironment();
        String namespace = StringUtils.ifEmpty(properties.getNamespace(), StringUtils.EMPTY);
        this.type = ObjectUtils.lookupSuperclassParameterized(clazz, RocketMQListener.class);
        this.topic = environment.resolvePlaceholders(namespace + annotation.topic());
        this.group = environment.resolvePlaceholders(namespace + annotation.group());
        this.quantity = annotation.quantity();
        this.duration = Duration.ofSeconds(annotation.duration());
        this.listener = listener;
        Duration interval = Duration.ofSeconds(annotation.interval());
        FilterExpression filter = new FilterExpression(annotation.tag(), FilterExpressionType.TAG);
        SimpleConsumerBuilder builder = provider.newSimpleConsumerBuilder().setClientConfiguration(configuration)
                .setConsumerGroup(this.group).setAwaitDuration(interval)
                .setSubscriptionExpressions(Collections.singletonMap(this.topic, filter));
        try {
            this.consumer = builder.build();
        } catch (ClientException e) {
            throw new RuntimeException(e);
        }

        // 启动消费线程
        this.thread = this.startup();
    }

    @Override
    public void destroy() throws Exception {
        this.running = false;
        try {
            this.shutdown(this.thread);
        } finally {
            this.shutdown(this.consumer);
        }
    }
}
